from typing import List
import  anxcor.abstractions as ab
import  anxcor.utils as os_utils
from  obspy.core import read, Stream, UTCDateTime
import xarray as xr
FLOAT_PRECISION = 1e-9
import numpy as np

def execute_if_ok_else_pass_through(method, one, two):
    if one is None and two is not None:
        return two
    elif one is not None and two is None:
        return one
    elif one is None and two is None:
        return None
    else:
        return method(one, two)

def method_per_op(method,opmethod, one, two):
    if one is None and two is not None:
        return opmethod(two)
    elif one is not None and two is None:
        return opmethod(one)
    elif one is None and two is None:
        return None
    else:
        return method(opmethod(one), opmethod(two))

class AnxcorDatabase:
    """ An interface for providing obspy.core.Stream objects to Anxcor objects

    get_stations() : List[str]
        returns a list of all stations in the dataset, formatted as a string with a network code
        e.g. 'Network.Station' ex: 'UU.1'

    get_waveforms(network : str =None, station : str =None,
                 starttime : float =None,endtime : =None,**kwargs) -> Stream:
        returns an obspy Stream of waveforms. Must take the kwargs:
        network and station as strings, and starttime, endtime as UTCDateTime timestamps
    """

    def __init__(self):
        pass

    def get_stations(self)-> List[str]:
        """
        returns a list of all stations in the dataset, formatted as a string with a network code
        e.g. 'Network.Station' ex: 'UU.1'

        Returns
        -------
        List[str]
            a list of strings representing stations. Must be formatted 'network_code.station_code'
        """
        raise NotImplementedError('Method: \'get_stations()\' method is not implemented')

    def get_waveforms(self,**kwarg_execute)->Stream:
        """
        returns an obspy Stream of waveforms

        Parameters
        ----------
        network : str,
            network to get waveform from. is called as a non-optional keyword-argument
        station : str,
            station to get waveform from. is called as a non-optional keyword-argument
        starttime : float,
            starttime of waveform in UTCDateTime timestamp. Non-optional keyword-argument
        endtime : float,
            endtime of waveform in UTCDateTime timestamp. Non-optional keyword-argument

        Returns
        -------
        Stream
            an obspy stream of traces

        Note
        ----
        Anxcor leaves both removing instrument response and ensuring continuity of data up to the user.
        If a given get_waveforms() query cannot deliver continuous data, it is perfectly fine to return a None object
        instead of an obspy Stream. Anxcor will handle the missing data accordingly.
        """
        raise NotImplementedError('Method: \'get_waveforms()\' is not implemented!')


class DataLoader(ab.XDatasetProcessor):

    def __init__(self, window_length):
        super().__init__()
        self._window_length = window_length
        self._datasets = {}

    def add_dataset(self, dataset: AnxcorDatabase, name: str, **kwargs):
        self._datasets[name]=dataset

    def get_stations(self) -> List[str]:
        """

        Returns
        -------
        station_list
            a list of all possible stations contained in the databases

        """
        station_list = []
        for key, value in self._datasets.items():
            one          = value
            seed_id_list = one.get_stations()
            for station in seed_id_list:
                if station not in station_list:
                    station_list.append(station)

        return station_list

    def has_data(self):
        return len(self._datasets.keys())>0

    def _combine(self, total_list, produced_list, key):
        for trace in produced_list:
            trace.stats.data_type = key
        total_list.extend(produced_list.traces)
        return total_list

    def _load_key(self,data_key,extension):
        load_key = '{}:{}@{}'.format('Trace IO',data_key,extension)
        return load_key

    def _response_mean_trend(self,data_key,extension):
        response_key = '{}:{}@{}'.format('RTrend, Mean, Resp',data_key,extension)
        return response_key

    def _stream_gather(self,data_key,extension):
        gather = '{}:{}@{}'.format('gather',data_key,extension)
        return gather

    def _single_thread_execute(self,*args, station=0, starttime=0, **kwargs):
        network, station = station.split('.')

        kwarg_execute    = {
            'network' : network,
            'station' : station,
            'starttime':starttime,
            'endtime':  starttime + self._window_length
            }

        traces = []
        for name, dataset in self._datasets.items():
            stream = dataset.get_waveforms(**kwarg_execute)
            stream = self._curate(stream,starttime)
            traces = self._combine(traces, stream, name)
        return Stream(traces=traces)

    def _io_result(self, result, starttime, source, format='mseed', **kwargs):
        type_dict = {}
        path      = None
        for trace in result:
            trace_id = trace.get_id()
            time     = trace.stats.starttime.isoformat()
            path     = os_utils.make_path_from_list([self._file, source, time])

            trace.write(path + os_utils.sep + trace_id + '.' + format, format=format)
            type_dict[trace_id]=trace.stats.data_type

        new_result = read(path + os_utils.sep + '*.' + format)
        pass_on = []
        for trace in new_result:
            trace_id = trace.get_id()
            if trace_id in type_dict:
                trace.stats.data_type = type_dict[trace_id]
                pass_on.append(trace)

        return pass_on

    def _format_path(self, extension, format, trace_id):
        path = self._format_folder_path(extension)
        path = path + trace_id + '.' + format
        return path

    def _format_folder_path(self,extension):
        path = self._file + os_utils.sep + extension + os_utils.sep
        return path

    def _metadata_to_persist(self, *param, **kwargs):
        return None

    def _get_name(self,*args):
        return None

    def _get_process(self):
        return 'load'

    def _curate(self, stream, starttime):
        endtime = starttime + self._window_length
        valid_traces = []
        for trace in stream:
            st = trace.stats.starttime.timestamp
            end= trace.stats.endtime.timestamp
            delta=trace.stats.delta

            cond1 = abs(st  - starttime)  <= delta + FLOAT_PRECISION
            cond2 = abs(end - endtime)    <= delta + FLOAT_PRECISION
            if hasattr(trace.data,'mask'):
                trace.data = trace.data.data
            if cond1 and cond2 and not np.any(np.isnan(trace.data)):
                valid_traces.append(trace)

        return Stream(traces=valid_traces)

    def _should_process(self, *args):
        return True

    def _window_key_convert(self,window):
        return UTCDateTime(int(window*100)/100).isoformat()

    def _addition_read_processing(self, result):
        name   = list(result.data_vars)[0]
        xarray       = result[name].copy()
        xarray.attrs = result.attrs.copy()
        del result
        return xarray



class XArrayCombine(ab.XDatasetProcessor):

    def __init__(self,**kwargs):
        super().__init__(**kwargs)

    def _single_thread_execute(self,first_data, second_data,**kwargs):
        if isinstance(first_data,xr.DataArray):
            first_data = first_data.to_dataset()
        if isinstance(second_data,xr.DataArray):
            second_data = second_data.to_dataset()
        return execute_if_ok_else_pass_through(self._normal_combine,first_data,second_data)

    def _normal_combine(self, first_data, second_data):
        if isinstance(first_data, xr.DataArray) and isinstance(second_data, xr.DataArray):
            if first_data.name == second_data.name:
                result = xr.concat([first_data, second_data], dim='pair')
            else:
                result = xr.merge([first_data, second_data])

        elif isinstance(first_data, xr.Dataset) and isinstance(second_data, xr.DataArray):
            result = self._merge_DataArray_Dataset(first_data, second_data)
        elif isinstance(first_data, xr.DataArray) and isinstance(second_data, xr.Dataset):
            result = self._merge_DataArray_Dataset(second_data, first_data)
        else:
            result = xr.merge([first_data, second_data])
        if not isinstance(result, xr.Dataset):
            name   = result.name
            coords = result.coords
            result = xr.Dataset(data_vars={name: result}, coords=coords)
        return result

    def _merge_DataArray_Dataset(self, data_set, data_array):
        if data_array.name in data_set.data_vars.keys():
            result = xr.concat([data_set[data_array.name], data_array], dim='pair')
            result = xr.Dataset(data_vars={data_array.name: result},coords=result.coords)
            result = data_set.combine_first(result)
        else:
            result = xr.merge([data_set, data_array])
        return result

    def _metadata_to_persist(self, first_data, second_data, **kwargs):
        if first_data is None and second_data is None:
            return None
        elif first_data is None and second_data is not None:
            return {**self._extract_metadata_dict(second_data)}
        elif first_data is not None and second_data is None:
            return {**self._extract_metadata_dict(first_data)}
        else:
            persist1 = self._extract_metadata_dict(first_data)
            persist2 = self._extract_metadata_dict(second_data)
        result   = {**persist1,**persist2}
        return result

    def _extract_metadata_dict(self, data_array):
        if isinstance(data_array, xr.DataArray) and len(list(data_array.coords['pair'].values))==1:
            key = list(data_array.coords['pair'].values)[0]
            dict1 = {key: {'stacks': data_array.attrs['stacks'],
                           'starttime': data_array.attrs['starttime'],
                           'endtime': data_array.attrs['endtime']},
                           'delta': data_array.attrs['delta'],
                           'operations': data_array.attrs['operations']}

            if 'location' in data_array.attrs.keys():
                dict1[key]['location']=data_array.attrs['location']
        else:
            dict1 = data_array.attrs

        return dict1

    def _io_result(self, result, *args, **kwargs):
        return result

    def _get_name(self, *args):
        return None

    def _get_process(self):
        return 'combine'

    def _window_key_convert(self,window):
        return window


    def starttime_parser(self,first,second):
        return 'depth:{}branch:{}'.format(first, second)



class XArrayStack(ab.XArrayProcessor):

    def __init__(self,**kwargs):
        super().__init__(**kwargs)


    def _single_thread_execute(self,first: xr.DataArray, second: xr.DataArray,*args,**kwargs):
        if first is None and second is not None:
            return second
        elif first is not None and second is None:
            return first
        elif first is None and second is None:
            return None
        else:
            result       = first + second
        return result


    def _metadata_to_persist(self, xarray_1,xarray_2,*args, **kwargs):
        return  method_per_op(self._combine_metadata,self._getattr,xarray_1,xarray_2)

    def _combine_metadata(self,attrs1, attrs2):
        attrs = {}
        attrs['delta'] = attrs1['delta']
        attrs['starttime'] = self._get_lower(attrs1, attrs2, 'starttime')
        attrs['endtime'] = self._get_upper(attrs1, attrs2, 'endtime')
        attrs['stacks'] = attrs1['stacks'] + attrs2['stacks']
        attrs['operations'] = attrs1['operations']
        if 'location' in attrs1.keys():
            attrs['location'] = attrs1['location']
        return attrs

    def _getattr(self,array):
        return array.attrs

    def _get_name(self,one,two):
        if one is not None:
            return one.name
        if two is not None:
            return two.name
        return None

    def _get_lower(self,one,two,key):
        if one[key] < two[key]:
            return one[key]
        return two[key]

    def _get_upper(self,one,two,key):
        if one[key] > two[key]:
            return one[key]
        return two[key]

    def _get_process(self):
        return 'stack'

    def _should_process(self,xarray1,xarray2, *args):
        return xarray1 is not None or xarray2 is not None


    def _window_key_convert(self,window):
        return window

    def starttime_parser(self,first,second):
        return 'depth:{}branch:{}'.format(first, second)








corr_pack = None


def import_numpy_backend():
    corr_pack = NumpyFFTContract()



class _FFtpack_Contract:


    def __init__(self):
        pass

    def crosscorrelate_two_arrays(self, xarray1, xarray2):
        raise NotImplementedError('please implement crosscorrelate_two_arrays()')


class NumpyFFTContract(_FFtpack_Contract):

    def __init__(self):
        super().__init__()
        import numpy as np
        self.np = np


    def crosscorrelate_two_arrays(self,xarray1, xarray2):
        src_chan_size = xarray1.data.shape[0]
        rec_chan_size = xarray2.data.shape[0]
        time_size     = xarray1.data.shape[-1]
        src_data      = xarray1.data.reshape(src_chan_size, time_size)
        receiver_data = xarray2.data.reshape(rec_chan_size, time_size)

        corr_length = time_size * 2 - 1
        target_length = self.np.fftpack.next_fast_len(corr_length)
        fft_src = self.np.conj(self.np.np.fft.rfft(src_data, target_length, axis=-1))
        fft_rec = self.np.fft.rfft(receiver_data, target_length, axis=-1)

        result = self._multiply_in_mat(fft_src, fft_rec)

        xcorr_mat = self.np.real(
            self.np.fft.fftshift(
            self.np.fft.irfft(result, corr_length, axis=-1), axes=-1)).astype(self.np.float64)
        return xcorr_mat

    def _multiply_in_mat(self,one, two):
        dtype = self.np.complex64
        zero_mat = self.np.zeros((one.shape[0],
                             two.shape[0],
                             one.shape[-1]), dtype=dtype)

        for ri in range(0, two.shape[0]):
            zero_mat[:, ri, :] = one

        for si in range(0, one.shape[0]):
            zero_mat[si, :, :] *= two

        return zero_mat


from  anxcor.containers import DataLoader, XArrayCombine, XArrayStack, AnxcorDatabase
from  anxcor.xarray_routines import XArrayConverter, XArrayResample, XArrayXCorrelate
from anxcor.abstractions import XArrayProcessor
from typing import List
import xarray as xr
import numpy as np
import itertools
from obspy.core import UTCDateTime, Stream, Trace
import json
import anxcor.utils as utils


class Anxcor:



    def __init__(self, window_length: float,**kwargs):
        """

        Parameters
        ----------
        window_length: float
            the window length in seconds
        """
        self._data      = _AnxcorData(window_length=window_length, **kwargs)
        self._processor = _AnxcorProcessor(self._data)
        self._converter = _AnxcorConverter(self._data)


    def add_dataset(self, database: AnxcorDatabase, name: str, **kwargs) -> None:
        """
        Adds a new dataset implementing the AnxorDatabase Interface to the crosscorrelation.
        If desired, you may provide an additional callable function used to remove the instrument
        response at every step.

        Parameters
        ----------
        database : AnxcorDatabase
            a database to add which implements the AnxorDatabase interface
        name : str
            a name describing the type of the resultant data
        trace_prep : Callable[[Trace],Trace], Optional
            an optional callable function to remove the instrument response.


        """
        self._data.add_dataset(database, name, **kwargs)


    def add_process(self, process: XArrayProcessor) -> None:
        """
        Add an XArrayProcessor to the processing steps. Functions are performed in the order they are added

        Parameters
        ----------
        process : XArrayProcessor
            an XArrayProcessor object

        """
        self._data.add_process(process)

    def set_task_kwargs(self, task: str, kwargs: dict):
        """
        sets keyword-arguments into the given task.

        Parameters
        ----------
        task : str
            the task key
        kwargs : dict
            the keyword-arguments to set

        """
        self._data.set_task_kwargs(task,kwargs)

    def set_process_kwargs(self, process: str, kwargs: dict):
        """
        sets keyword-arguments into the given process

        Parameters
        ----------
        process : str
            the process key
        kwargs : dict
            the keyword-arguments to set

        """
        self._data.set_process_kwargs(process,kwargs)

    def get_task(self,key: str) -> XArrayProcessor:
        """
        Returns an XArrayProcessor task based on its key

        Parameters
        ----------
        key : str
            the key to the processor object

        Returns
        -------
        XArrayProcessor, List[XArrayProcessor]
            either the processor or the list of processes defined by the task key

        """
        return self._data.get_task(key)

    def get_starttimes(self,starttime: float, endtime: float, overlap: float) -> List[float]:
        """
        Get a list of starttimes based on a given starttime, endtime, and percent overlap of windows

        Parameters
        ----------
        starttime : float
            the starttime (as a UTCDateTime timestamp)
        endtime : float
            the endtime (as a UTCDateTime timestamp)
        overlap : float
            percent overlap of windows as a decimal (100% = 1.0). Must be between 0 - 1.

        Returns
        -------
        List[float]
            a list of UTCDateTime timestamps representing window starttimes

        """
        return self._converter.get_starttimes(starttime, endtime, overlap)

    def save_at_task(self, folder: str,task,**kwargs) -> None:
        """
        Save result of method to file at a given task step

        Parameters
        ----------
        folder : str
            the folder to save the step result to.
        task : str
            one of:
            'data'
            'resample'
            'xconvert'
            'correlate'
            'stack'
            'combine'
        representing the process key at which to save.

        Note
        ----
        see save_at_process() for saving after a specific processing step

        """
        self._data.save_at_task(folder,task)

    def load_at_task(self, folder: str,task,**kwargs)-> None:
        """
        load result of method from file at a given task step.
        will disable all routines behind it in the compute graph

        Parameters
        ----------
        folder : str
            the folder to save the step result to.
        task : str
            one of:
            'data'
            'resample'
            'xconvert'
            'correlate'
            'stack'
            'combine'
        representing the process key at which to save.

        Note
        ----
        see load_at_process() for saving after a specific processing step

        """
        self._data.load_at_task(folder, task)

    def save_at_process(self, folder: str,process : str,**kwargs) -> None:
        """
        save result of method to file at a given task step

        Parameters
        ----------
        folder : str
            the folder to save the step result to.
        process : str
            a string returned by the get_name()/get_process() method of an XArrayProcessor
            object assigned to an Anxcor object

        Note
        ----
        the object must have already been added via the add_process() routine

        """
        self._data.save_at_process(folder,process)

    def load_at_process(self, folder: str,process: str,**kwargs)-> None:
        """
        load result of method from file at a given task step
        will disable all routines behind it in the compute graph

        Parameters
        ----------
        folder : str
            the folder to save the step result to.
        process : str
            a string returned by the get_name()/get_process() method of an XArrayProcessor
            object assigned to an Anxcor object

        Note
        ----
        the object must have already been added via the add_process() routine

        """
        self._data.load_at_process(folder, process)

    def process(self, starttimes: List[float], dask_client=None) -> xr.DataSet:
        """
        Perform the crosscorrelation routines. By default this processes, stacks, and combines
        all crosscorrelations during the given starttimes. See documentation for finer control

        Parameters
        ----------
        starttimes : List[float]
            a list of UTCDateTime timestamps representing window starttimes
        dask_client : Optional
            an optional dask_client instance for parallel processing using dask

        Returns
        -------
        Dataset, Future,

            if single threaded, will return a single XArray Dataset object containing the
        stacked crosscorrelations. If a dask client is provided it will return a future instance.
        See Dask Documentation for further details
        """
        if self._data.has_data():
            return self._processor.process(starttimes, dask_client=dask_client)
        else:
            print('no data banks added. will not execute.')

    def xarray_to_obspy(self, xdataset: xr.DataSet):
        """
        convert the output of a anxcor correlation into an obspy stream

        Parameters
        ----------
        xdataset : Dataset
            an xarray Dataset object produced by AnXcor

        Returns
        -------
        Stream
            obspy trace stream of cross correlations

        """
        return self._converter.xarray_to_obspy(xdataset)

    def save_config(self,file):
        """
        saves the parameters for each processing step as a .ini file
        Parameters
        ----------
        file : str
            the config file to save
        """
        self._data.save_config(file)

    def load_config(self,file):
        """
        loads a previously built .ini file, assigning the results to the current process stack
        will present a warning if config file parameters are not present
        Parameters
        ----------
        file : str
            the config file to load


        """
        self._data.load_config(file)



class _AnxcorProcessor:

    time_format = '%d-%m-%Y %H:%M:%S'
    CORR_FORMAT = 'src:{} rec:{}'
    def __init__(self,data):
        self.data =data

    def _station_window_operations(self, channels, dask_client=None, starttime=None, station=None):
        xarray      = self.data.get_task('xconvert')(channels, starttime=starttime, station=station, dask_client=dask_client )
        downsampled = self.data.get_task('resample')(xarray, starttime=starttime, station=station, dask_client=dask_client )
        tasks = [downsampled]
        process_list = self.data.get_process_order()
        for process_key in process_list:
            process = self.data.get_process(process_key)
            task   = tasks.pop()
            result = process(task,starttime=starttime, station=station, dask_client=dask_client )
            tasks.append(result)
        return tasks[0]


    def process(self,starttimes, dask_client=None,**kwargs):

        station_pairs = self.data.get_station_combinations()
        futures = []
        for pair in station_pairs:

            correlation_list  = self._iterate_starttimes(pair,  starttimes,dask_client=dask_client)
            correlation_stack = self._reduce(correlation_list,
                                             station=str(pair),
                                             reducing_func=self.data.get_task('stack'),
                                             dask_client=dask_client)

            futures.append(correlation_stack)

        combined_crosscorrelations = self._reduce(futures,
                                                  station='combine',
                                                  reducing_func=self.data.get_task('combine'),
                                                  dask_client=dask_client)

        return combined_crosscorrelations


    def _iterate_starttimes(self, pair, starttimes, dask_client=None):
        source   = pair[0]
        receiver = pair[1]
        correlation_stack = []
        for starttime in starttimes:
            source_channels = self.data.get_task('data')(
                                                  starttime=starttime,
                                                  station=source,
                                                  dask_client=dask_client)
            source_ch_ops = self._station_window_operations(source_channels,
                                                            starttime=starttime,
                                                            station=source,
                                                            dask_client=dask_client)
            if source==receiver:
                receiver_ch_ops = source_ch_ops
            else:
                receiver_channels = self.data.get_task('data')(
                                                      starttime=starttime,
                                                      station=receiver,
                                                      dask_client=dask_client)
                receiver_ch_ops   = self._station_window_operations(receiver_channels,
                                                                starttime=starttime,
                                                                station=receiver,
                                                                dask_client=dask_client)

            correlation = self.data.get_task('crosscorrelate')(source_ch_ops,
                                                   receiver_ch_ops,
                                                   station=self.CORR_FORMAT.format(source,receiver),
                                                   starttime=starttime,
                                                   dask_client=dask_client)

            correlation_stack.append(correlation)

        return correlation_stack

    def _reduce(self,
                future_stack,
                station = None,
                reducing_func = None,
                dask_client=None):

        tree_depth = 1
        while len(future_stack) > 1:
            new_future_list = []

            while len(future_stack) > 1:
                branch_index = len(future_stack)
                first  = future_stack.pop()
                second = future_stack.pop()
                result = reducing_func(first, second,
                                       station=station,
                                       starttime=reducing_func.starttime_parser(tree_depth, branch_index),
                                       dask_client=dask_client)
                new_future_list.append(result)

            if len(future_stack) == 1:
                branch_index = len(future_stack)
                first  = future_stack.pop()
                second = new_future_list.pop()
                result = reducing_func(first, second,
                                       station=station,
                                       starttime=reducing_func.starttime_parser(tree_depth, branch_index),
                                       dask_client=dask_client)
                new_future_list.append(result)
            tree_depth +=1
            future_stack=new_future_list

        result = reducing_func(future_stack[0], None,
                      station=station,
                      starttime=reducing_func.starttime_parser('Final', 0),
                      dask_client=dask_client)
        return result



class _AnxcorData:
    TASKS = ['data', 'xconvert', 'resample', 'process', 'crosscorrelate', 'stack', 'combine']
    def __init__(self,window_length: float =3600.0):
        self._window_length=window_length
        self._tasks = {
            'data': DataLoader(window_length),
            'xconvert': XArrayConverter(),
            'resample': XArrayResample(),
            'process' : {},
            'crosscorrelate': XArrayXCorrelate(),
            'combine': XArrayCombine(),
            'stack': XArrayStack()
            }
        self._process_order = []

    def get_process_order(self):
        return self._process_order.copy()

    def get_process(self,process_key):
        return self._tasks['process'][process_key]

    def get_window_length(self):
        return self._window_length

    def get_station_combinations(self):
        stations = self._tasks['data'].get_stations()
        station_pairs = list(itertools.combinations_with_replacement(stations, 2))
        return station_pairs

    def has_data(self):
        return self._tasks['data'].has_data()

    def save_at_task(self, folder, task : str='resample'):
        if task in self._tasks.keys() and task!='process':
             self._tasks[task].set_io_task(folder, 'save')
        else:
            print('{} is Not a valid task'.format(task))

    def save_at_process(self, folder, process : str='whiten'):
        if process in self._process_order:
             self._tasks['process'][process].set_io_task(folder, 'save')
        else:
            print('{} is Not a valid process'.format(process))


    def load_at_process(self, folder, process='resample'):
        if process in self._process_order:
            self._tasks['process'][process].set_io_task(folder, 'load')
            self._disable_tasks('process')
            self._disable_process(process)
        else:
            print('{} is Not a valid process'.format(process))

    def load_at_task(self, folder, task='resample'):
        if task in self._tasks.keys() and task!='process':
             self._tasks[task].set_io_task(folder, 'load')
             self._disable_tasks(task)
        else:
            print('{} is Not a valid task'.format(task))


    def _disable_process(self, process):
        index = self._process_order.index(process)
        for disable in range(0, index + 1):
            pr = self._process_order[disable]
            self._tasks['process'][pr].disable()


    def _disable_tasks(self, key):
        index = self.TASKS.index(key)
        for disable in range(0, index + 1):
            task = self.TASKS[disable]
            if task is not 'process':
                self._tasks[task].disable()
            else:
                if self._process_order:
                    self._disable_process(self._process_order[-1])


    def add_dataset(self, dataset, name, trace_prep=None, **kwargs):
        self._tasks['data'].add_dataset(dataset, name, trace_prep=trace_prep, **kwargs)


    def add_process(self, process):
        key = process.get_name()
        self._tasks['process'][key]=process
        self._process_order.append(key)

    def set_task_kwargs(self,task: str,kwargs: dict):
        if task in self._tasks.keys():
            self._tasks[task].set_kwargs(kwargs)
        else:
            print('{}: is not a valid task. ignoring'.format(task))

    def set_process_kwargs(self,process: str, kwargs: dict):
        if process in self._tasks['process'].keys():
            self._tasks['process'][process].set_kwargs(kwargs)
        else:
            print('{}: is not a valid process. ignoring'.format(process))

    def get_task(self,key):
        if key in self._tasks.keys():
            return self._tasks[key]
        else:
            raise KeyError('key does not exist in tasks')

    def save_config(self, file):
        config = {}
        for task in self.TASKS:
            if task!='process':
                config[task] = self._tasks[task].get_kwargs()
            else:
                config[task] = {
                    'processing order' : self._process_order,
                }
                for key, process in self._tasks[task].items():
                    config[task][key]=process.get_kwargs()

        folder = utils.get_folderpath(file)
        if not utils.folder_exists(folder):
            print('given folder to save file does not exist. ignoring save call')
            return None
        with open(file, 'w') as configfile:
            json.dump(config,configfile, sort_keys=True, indent=4)


    def load_config(self, file):
        if not utils.file_exists(file):
            print('given .ini file to load file does not exist. Inoring load call')
            return None

        with open(file, 'r') as p_file:
            config = json.load(p_file)
        for task in self.TASKS:
            if task!='process':
                self._tasks[task].set_kwargs(config[task])
            else:
                self._process_order = config[task]['processing order']
                for key, process_kwargs in config[task].items():
                    if key!='processing order':
                        if key not in self._tasks[task].keys():
                            print('{} does not exist inside current Anxcor Object\n'.format(key)+\
                                  'Skipping for now..')
                        else:
                            self._tasks[task][key].set_kwargs(process_kwargs)


class _AnxcorConverter:


    def __init__(self,data):
        self.data = data

    def get_starttimes(self, starttime, endtime, overlap):
        starttimes = []
        delta      = self.data.get_window_length() * (1 - overlap)
        time = starttime
        while time < endtime:
            starttimes.append(time)
            time+=delta
        return starttimes

    def xarray_3D_to_2D(self,xdataset: xr.DataSet):
        # get vars
        new_ds = xdataset.assign_coords(component='{}{}{}'.format(xdataset.src_chan,':',xdataset.rec_chan))
        new_ds = new_ds.drop_dims(['src_chan','rec_chan'])
        return new_ds

    def xarray_2D_to_3D(self,xdataset: xr.DataSet):
        new_ds = xdataset.assign_coords(src_chan=xdataset.component.split(':')[0])
        new_ds = new_ds.assign_coords(rec_chan=xdataset.component.split(':')[1])
        new_ds = new_ds.drop_dims(['component'])
        return new_ds

    def xarray_to_obspy(self,xdataset: xr.DataSet):
        attrs = xdataset.attrs
        traces = []
        starttime = list(xdataset.coords['time'].values)[0]
        starttime = self._extract_timestamp(starttime)
        for name in xdataset.data_vars:
            xarray = xdataset[name]
            for pair in list(xdataset.coords['pair'].values):
                for src_chan in list(xdataset.coords['src_chan'].values):
                    for rec_chan in list(xdataset.coords['rec_chan'].values):

                        record = xarray.loc[dict(pair=pair,src_chan=src_chan,rec_chan=rec_chan)]

                        station_1, station_2, network_1, network_2 = self._extract_station_network_info(pair)
                        header_dict = {
                        'delta'    : record.attrs['delta'],
                        'npts'     : record.data.shape[-1],
                        'starttime': starttime,
                        'station'  : '{}.{}'.format(station_1,station_2),
                        'channel'  : '{}.{}'.format(src_chan,rec_chan),
                        'network'  : '{}.{}'.format(network_1,network_2)
                        }
                        trace = Trace(data=record.data,header=header_dict)
                        traces.append(trace)

        return Stream(traces=traces)

    def _extract_timestamp(self,starttimestamp):
        starttimestamp = starttimestamp.astype(np.float64)/1e9
        return UTCDateTime(starttimestamp)

    def _extract_station_network_info(self,pair):
        stations = [0, 0]
        pairs = pair.split('rec:')
        stations[0] = pairs[0].split(':')[1]
        stations[1] = pairs[1]

        station_1 = stations[0].split('.')[1]
        station_2 = stations[1].split('.')[1]
        network_1 = stations[0].split('.')[0]
        network_2 = stations[1].split('.')[0]
        return station_1, station_2, network_1, network_2

    def align_station_pairs(self,xdataset: xr.DataSet):
        attrs = xdataset.attrs.copy()
        del attrs['delta']
        del attrs['operations']
        print(attrs.keys())
        for name in xdataset.data_vars:
            xarray = xdataset[name]


"""
xarray.DataArray operations for use with Anxcor processing routines

"""

import numpy as np
import xarray as xr
import anxcor.filters as filt_ops
import pandas as pd
import  anxcor.abstractions as ab
from obspy.core import UTCDateTime
OPERATIONS_SEPARATION_CHARACTER = '\n'
SECONDS_2_NANOSECONDS = 1e9


class XArrayConverter(ab.XArrayProcessor):
    """
    converts an obspy stream into an xarray

    """

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.writer = ab._XArrayWrite(None)
        self.reader = ab._XArrayRead(None)

    def _create_xarray_dataset(self, trace_list):
        return None

    def _attach_metadata(stream, xarr):
        return None

    def _get_station_id(self,trace):
        network = trace.stats.network
        station = trace.stats.station
        return network + '.' + station

    def _single_thread_execute(self, stream,*args,starttime=0, **kwargs):
        if stream is not None and len(stream)>0:
            return self._convert_trace_2_xarray(stream)
        return None

    def _convert_trace_2_xarray(self, stream):
        channels = []
        stations = []
        delta = None
        time_array = None
        data_type  = 'default'
        starttime  = None
        latitude   = None
        longitude  = None
        elevation  = None
        for trace in stream:
            station_code = self._get_station_id(trace)
            if station_code not in stations:
                stations.append(station_code)

            channel = trace.stats.channel
            if channel not in channels:
                channels.append(channel)
            if time_array is None:
                starttime = np.datetime64(trace.stats.starttime.datetime)
                endtime = np.datetime64(trace.stats.endtime.datetime)

                delta = trace.stats.delta
                timedelta = pd.Timedelta(delta, 's').to_timedelta64()

                time_array = np.arange(starttime, endtime + timedelta, timedelta)
                starttime = trace.stats.starttime.timestamp
                if hasattr(trace.stats,'data_type'):
                    data_type = trace.stats.data_type
                stats_keys = trace.stats.keys()
                if 'latitude' in stats_keys:
                    latitude = trace.stats.latitude
                if 'longitude' in stats_keys:
                    longitude = trace.stats.longitude
                if 'elevation' in stats_keys:
                    elevation = trace.stats.elevation
        empty_array = np.zeros((len(channels), len(stations), len(time_array)))
        for trace in stream:
            chan = channels.index(trace.stats.channel)
            station_id = stations.index(self._get_station_id(trace))

            empty_array[chan, station_id, :] = trace.data
        return self._create_xarray(channels, data_type, delta, elevation, empty_array, latitude, longitude, starttime,
                                   stations, time_array)

    def _create_xarray(self, channels, data_type, delta, elevation, empty_array, latitude, longitude, starttime,
                       stations, time_array):
        xarray = xr.DataArray(empty_array, coords=[channels, stations, time_array],
                              dims=['channel', 'station_id', 'time'])
        xarray.name = data_type
        xarray.attrs['delta'] = delta
        xarray.attrs['starttime'] = starttime
        xarray.attrs['operations'] = 'xconvert'
        if latitude is not None and longitude is not None:
            if elevation is not None:
                xarray.attrs['location'] = (latitude, longitude, elevation)
            else:
                xarray.attrs['location'] = (latitude, longitude)
        return xarray

    def _metadata_to_persist(self, *param, **kwargs):
        return None

    def _get_process(self):
        return 'xconvert'

    def _get_name(self,*args):
        return None


class XArrayBandpass(ab.XArrayProcessor):
    """
    applies a bandpass filter to a provided xarray
    """

    def __init__(self,upper_frequency=10.0,lower_frequency=0.01,order=2,taper=0.1,**kwargs):
        super().__init__(**kwargs)
        self._kwargs = {'upper_frequency':upper_frequency,
                        'lower_frequency':lower_frequency,
                        'order':order,
                        'taper':taper}

    def _single_thread_execute(self, xarray: xr.DataArray,*args, **kwargs):
        sampling_rate = 1.0 / xarray.attrs['delta']
        ufunc_kwargs = {**self._kwargs}

        if self._kwargs['upper_frequency'] > sampling_rate / 2:
            ufunc_kwargs['upper_frequency'] = sampling_rate / 2

        filtered_array = xr.apply_ufunc(filt_ops.taper, xarray,
                                        input_core_dims=[['time']],
                                        output_core_dims=[['time']],
                                        kwargs={**self._kwargs})
        filtered_array = xr.apply_ufunc(filt_ops.butter_bandpass_filter, tapered_array,
                                        input_core_dims=[['time']],
                                        output_core_dims=[['time']],
                                        kwargs={**ufunc_kwargs,**{
                                                'sample_rate': sampling_rate}})

        return filtered_array

    def _add_operation_string(self):
        return 'bandpass@{}<{}'.format(self._kwargs['lower_frequency'],
                                       self._kwargs['upper_frequency'])

    def _get_process(self):
        return 'bandpass'


class XArrayTaper(ab.XArrayProcessor):
    """
    tapers signals on an xarray timeseries

    Note
    --------
    most XArrayProcessors which operate in the frequency domain
    have tapering as part of the process.

    """

    def __init__(self,taper=0.1,**kwargs):
        super().__init__(**kwargs)
        self._kwargs = {'taper':taper}

    def _single_thread_execute(self, xarray: xr.DataArray,*args, **kwargs):
        filtered_array = xr.apply_ufunc(filt_ops.taper, xarray,
                                        input_core_dims=[['time']],
                                        output_core_dims=[['time']],
                                        kwargs={**self._kwargs})

        return filtered_array

    def _add_operation_string(self):
        return 'taper@{}%'.format(self._kwargs['taper']*100)

    def _get_process(self):
        return 'taper'


class XArrayResample(ab.XArrayProcessor):
    """
    resamples the provided xarray to a lower frequency
    """

    def __init__(self, target_rate=10.0,**kwargs):
        super().__init__(**kwargs)
        self._kwargs['target_rate'] = target_rate

    def _single_thread_execute(self, xarray: xr.DataArray,*args,starttime=0,**kwargs):
        delta =  xarray.attrs['delta']
        sampling_rate = 1.0 / delta
        nyquist       = self._kwargs['target_rate'] / 2.0
        target_rule = str(int((1.0 /self._kwargs['target_rate']) * SECONDS_2_NANOSECONDS)) + 'N'

        mean_array     = xarray.mean(dim=['time'])
        demeaned_array = xarray - mean_array
        detrend_array  = xr.apply_ufunc(filt_ops.detrend, demeaned_array,
                                        input_core_dims=[['time']],
                                        output_core_dims = [['time']],
                                        kwargs={'type':'linear'})
        tapered_array = xr.apply_ufunc(filt_ops.taper, detrend_array,
                                        input_core_dims=[['time']],
                                        output_core_dims=[['time']],
                                        kwargs={**self._kwargs})
        filtered_array = xr.apply_ufunc(filt_ops.lowpass_filter,tapered_array,
                                        input_core_dims=[['time']],
                                        output_core_dims=[['time']],
                                        kwargs={'upper_frequency':    nyquist,
                                                'sampling_rate': sampling_rate})

        resampled_array= filtered_array.resample(time=target_rule)\
            .interpolate('linear')

        starttime_datetime = np.datetime64(UTCDateTime(starttime).datetime)
        endtime_datetime   = np.datetime64(UTCDateTime(starttime +
                                                       delta*resampled_array.data.shape[-1]).datetime)
        timedelta = pd.Timedelta(delta, 's').to_timedelta64()
        time_array = np.arange(starttime_datetime, endtime_datetime + timedelta, timedelta)
        resampled_array = resampled_array.interp(time=time_array).bfill('time').ffill('time')


        return resampled_array

    def _add_metadata_key(self):
        return ('delta',1.0/self._kwargs['target_rate'])

    def _get_process(self):
        return 'resample'

    def _add_operation_string(self):
        return 'resampled@{}Hz'.format(self._kwargs['target_rate'])



class XArrayXCorrelate(ab.XArrayProcessor):
    """
    correlates two xarrays channel-wise in the frequency domain.

    """

    def __init__(self,max_tau_shift=10.0,taper=0.01,**kwargs):
        super().__init__(**kwargs)
        self._kwargs['max_tau_shift']=max_tau_shift
        self._kwargs['taper'] = taper

    def _single_thread_execute(self, source_xarray: xr.DataArray, receiver_xarray: xr.DataArray,*args, **kwargs):
        if source_xarray is not None and receiver_xarray is not None:
            correlation = filt_ops.xarray_crosscorrelate(source_xarray,
                                             receiver_xarray,
                                                     **self._kwargs)
            return correlation
        return None

    def _get_process(self):
        return 'crosscorrelate'


    def _metadata_to_persist(self, xarray_1,xarray_2, **kwargs):
        if xarray_2 is None or xarray_1 is None:
            return None

        xarray_1 = xr.apply_ufunc(filt_ops.taper, xarray_1,
                                  input_core_dims=[['time']],
                                  output_core_dims=[['time']],
                                  kwargs={**self._kwargs})
        xarray_2 = xr.apply_ufunc(filt_ops.taper, xarray_2,
                                  input_core_dims=[['time']],
                                  output_core_dims=[['time']],
                                  kwargs={**self._kwargs})

        attrs = {'delta'    : xarray_1.attrs['delta'],
                 'starttime': xarray_1.attrs['starttime'],
                 'stacks'   : 1,
                 'endtime'  : xarray_1.attrs['starttime'] + xarray_1.attrs['delta'] * xarray_1.data.shape[-1],
                 'operations': xarray_1.attrs['operations'] + OPERATIONS_SEPARATION_CHARACTER + \
                               'correlated@{}<t<{}'.format(self._kwargs['max_tau_shift'],self._kwargs['max_tau_shift'])}
        if 'location' in xarray_1.attrs.keys() and 'location' in xarray_2.attrs.keys():
            attrs['location'] = {'src':xarray_1.attrs['location'],
                                 'rec':xarray_2.attrs['location']}
        return attrs

    def _add_operation_string(self):
        return None

    def  _should_process(self,xarray1, xarray2, *args):
        return xarray1 is not None and xarray2 is not None

    def _get_name(self,one,two):
        if one is not None and two is not None:
            return 'src:{} rec:{}'.format(one.name, two.name)
        else:
            return None

class XArrayRemoveMeanTrend(ab.XArrayProcessor):
    """
    removes the mean and trend of an xarray timeseries
    """
    def __init__(self,**kwargs):
        super().__init__(**kwargs)


    def _single_thread_execute(self,xarray,*args,**kwargs):

        mean_array = xarray.mean(dim=['time'])
        demeaned_array = xarray - mean_array
        detrend_array = xr.apply_ufunc(filt_ops.detrend, demeaned_array,
                                       input_core_dims=[['time']],
                                       output_core_dims=[['time']],
                                       kwargs={'type': 'linear'},keep_attrs=True)

        return detrend_array

    def _add_operation_string(self):
        return 'remove_Mean&Trend'

    def _get_process(self):
        return 'remove_mean_trend'


class XArrayTemporalNorm(ab.XArrayProcessor):
    """
    applies a temporal norm operation to an xarray timeseries
    """

    def __init__(self, time_window=10.0, lower_frequency=0.001,
                 upper_frequency=5.0, t_norm_type='reduce_channel',
                 rolling_procedure='mean',
                 reduction_procedure='max', **kwargs):
        super().__init__(**kwargs)
        self._kwargs['t_norm_type']         = t_norm_type
        self._kwargs['time_window']         = time_window
        self._kwargs['lower_frequency']     = lower_frequency
        self._kwargs['upper_frequency']     = upper_frequency
        self._kwargs['rolling_procedure']   = rolling_procedure
        self._kwargs['reduction_procedure'] = reduction_procedure


    def _single_thread_execute(self, xarray: xr.DataArray,*args, **kwargs):
        time_average         = self._kwargs['time_window']

        sampling_rate   = 1.0/xarray.attrs['delta']

        bandpassed_array = xr.apply_ufunc(filt_ops.butter_bandpass_filter,xarray,
                                        input_core_dims=[['time']],
                                        output_core_dims = [['time']],
                                        kwargs={**self._kwargs,
                                                **{'sample_rate':sampling_rate}})

        bandpassed_array = self._apply_rolling_method(bandpassed_array, sampling_rate)
        bandpassed_array = bandpassed_array.ffill('time').bfill('time')
        bandpassed_array = self._reduce_by_channel(bandpassed_array)


        xarray = xarray / bandpassed_array
        return xarray

    def _apply_rolling_method(self, bandpassed_array, sampling_rate):
        time_window       = self._kwargs['time_window']
        rolling_samples   = int(sampling_rate * time_window)
        rolling_procedure = self._kwargs['rolling_procedure']

        if rolling_procedure == 'mean':
            bandpassed_array = abs(bandpassed_array).rolling(time=rolling_samples,
                                                             min_periods=1,center=True).mean()
        elif rolling_procedure == 'median':
            bandpassed_array = abs(bandpassed_array).rolling(time=rolling_samples,
                                                             min_periods=1,center=True).median()
        elif rolling_procedure == 'min':
            bandpassed_array = abs(bandpassed_array).rolling(time=rolling_samples,
                                                             min_periods=1,center=True).min()
        elif rolling_procedure == 'max':
            bandpassed_array = abs(bandpassed_array).rolling(time=rolling_samples,
                                                             min_periods=1,center=True).max()
        else:
            bandpassed_array = abs(bandpassed_array)
        return bandpassed_array

    def _reduce_by_channel(self, bandpassed_array):
        reduction_procedure = self._kwargs['reduction_procedure']
        norm_type           = self._kwargs['t_norm_type']
        if norm_type == 'reduce_channel':
            if reduction_procedure   == 'mean':
                bandpassed_array = abs(bandpassed_array).mean(dim='channel')
            elif reduction_procedure == 'median':
                bandpassed_array = abs(bandpassed_array).median(dim='channel')
            elif reduction_procedure == 'min':
                bandpassed_array = abs(bandpassed_array).min(dim='channel')
            elif reduction_procedure == 'max':
                bandpassed_array = abs(bandpassed_array).max(dim='channel')
        return bandpassed_array

    def _get_process(self):
        return 'temp_norm'


    def _add_operation_string(self):
        return 'temporal_norm@type({}),rolling({}),reduce_by({}).'+\
               'time_mean({}),f_norm_basis({})<f<({})'.format(
            self._kwargs['t_norm_type'],self._kwargs['time_window'],
            self._kwargs['rolling_procedure'], self._kwargs['reduction_procedure'],
            self._kwargs['lower_frequency'],self._kwargs['upper_frequency'])



class XArrayWhiten(ab.XArrayProcessor):
    """
    whitens the frequency spectrum of a given xarray
    """
    def __init__(self, smoothing_window_ratio=10.0, lower_frequency=0.001,
                 upper_frequency=5.0, order=2,whiten_type='cross_component', **kwargs):
        super().__init__(**kwargs)
        self._kwargs = {
            'smoothing_window_ratio': smoothing_window_ratio,
            'lower_frequency': lower_frequency,
            'upper_frequency': upper_frequency,
            'order':order,
            'whiten_type':whiten_type}

    def _single_thread_execute(self, xarray: xr.DataArray,*args, **kwargs):
        channel   = xarray.get_axis_num(dim='channel')
        new_array = xr.apply_ufunc(filt_ops.xarray_whiten, xarray,
                                          input_core_dims=[['time']],
                                          output_core_dims=[['time']],
                                          kwargs={**self._kwargs ,
                                                  **{'delta':xarray.attrs['delta'],
                                                     'channel_axis':channel,
                                                     'axis' :xarray.get_axis_num('time')}})

        return  new_array

    def _get_process(self):
        return 'whiten'

    def _add_operation_string(self):
        return 'f_whiten@window_ratio({}),frequency_window({})<f<({})hz'.format( \
            self._kwargs['smoothing_window_ratio'],self._kwargs['lower_frequency'],
            self._kwargs['upper_frequency'])









from scipy.signal import butter, filtfilt, freqz, detrend, convolve
import scipy.fftpack as fftpack
import xarray as xr
from obspy.core import UTCDateTime
import numpy as np
import pandas as pd
STARTTIME_NS_PRECISION = 100.0
DELTA_MS_PRECISION     = 100.0/1


def _butter_lowpass(cutoff, fs, order=5,**kwargs):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    return b, a

def lowpass_filter(data, upper_frequency=0.5, sampling_rate=1, order=2,axis=-1):
    """
     lowpass filter for n-dimensional arrays
    Parameters
    ----------
    data: np.ndarray
        n-d numpy array
    cutoff:
        upper cutoff frequency to filter
    fs:
        sampling rate
    order:
        order of the filter. defaults to 2
    axis:
        axis to apply the filter to

    Returns
    -------
        a filtered nd array of same dimensions as the input data
    """
    b, a = _butter_lowpass(upper_frequency, sampling_rate, order=order)
    y = filtfilt(b, a, data,axis=axis)
    return y

def _butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.500000000001 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


def butter_bandpass_filter(data, lower_frequency=0.01, upper_frequency=1.0, sample_rate=0.5, order=2,**kwargs):
    b, a = _butter_bandpass(lower_frequency, upper_frequency, sample_rate, order=order)
    y = filtfilt(b, a, data)
    return y

def _taper(np_arr,taper=0.1,**kwargs):
    time_length = np_arr.shape[0]
    taper_half_length = int(taper*time_length/2)

    full_window = np.hamming(int(taper*time_length))
    ones        = np.ones(time_length)
    ones[ :taper_half_length]*=full_window[:taper_half_length]
    ones[-taper_half_length:]*=full_window[-taper_half_length:]

    result = np.multiply(np_arr,ones)

    return result

def taper(data, taper=0.1,axis=-1,**kwargs):
    time_length = data.shape[axis]
    taper_half_length = int(taper*time_length/2)

    full_window = np.hamming(int(taper*time_length))
    ones        = np.ones(time_length)
    ones[ :taper_half_length]*=full_window[:taper_half_length]
    ones[-taper_half_length:]*=full_window[-taper_half_length:]

    result = data * ones

    return result

def xarray_whiten(data, taper=0.1, smoothing_window_ratio=0.1,
                        lower_frequency=None,
                        upper_frequency=None, axis=-1, delta=0.1,
                        whiten_type=None,channel_axis=0,**kwargs):

    len_time_axis = data.shape[axis]

    source_array  = np.apply_along_axis(_taper,axis,data,taper=taper)
    freq_domain   = _into_frequency_domain(source_array,axis=axis)

    nyquist = 0.5/delta
    if upper_frequency > nyquist:
        upper_frequency=nyquist

    convolve_window = int(smoothing_window_ratio*source_array.shape[axis])
    convolve_ones = np.ones((convolve_window,)) / convolve_window
    running_spec  = np.apply_along_axis(convolve, axis, np.abs(freq_domain), convolve_ones, mode='same')
    if whiten_type=='cross_component':
        running_spec = np.max(running_spec,axis=channel_axis)
    freq_domain   = np.divide(freq_domain, running_spec)

    b, a     = _butter_bandpass(lower_frequency, upper_frequency, 1 / delta, **kwargs)
    delta_fs = _get_deltaf(freq_domain, delta, axis=axis) * 2 * np.pi
    w, resp  = freqz(b, a, worN=delta_fs)
    freq_domain  *= resp
    rsult         = _into_time_domain(freq_domain,axis=axis)[:,:,:len_time_axis]
    time_domain   = np.real(fftpack.ifftshift(rsult,axes=axis))
    return time_domain


def _check_if_inputs_make_sense(source_array,  max_tau_shift):
    time = source_array.attrs['delta'] * (source_array.data.shape[2]-1)
    total_time = time
    if max_tau_shift is not None and total_time <= max_tau_shift:
        raise Exception('given tau shift is too large for input array')


def _check_xcorrelate_assumptions(source_xarray, receiver_xarray, taper, max_tau_shift, dummy_task, **kwargs):
    assert int(source_xarray.attrs['delta']   * DELTA_MS_PRECISION)/DELTA_MS_PRECISION == \
           int(receiver_xarray.attrs['delta'] * DELTA_MS_PRECISION)/DELTA_MS_PRECISION, \
                'array deltas are not equal!!'
    assert int(source_xarray.attrs['starttime']   * STARTTIME_NS_PRECISION)/STARTTIME_NS_PRECISION==\
           int(receiver_xarray.attrs['starttime'] * STARTTIME_NS_PRECISION)/STARTTIME_NS_PRECISION, \
                'differring starttimes!!'+\
                                             ' will not correlate'
    assert source_xarray.data.shape[-1]==receiver_xarray.data.shape[-1], \
        'xarray shapes are different! will not proceed'

def _will_not_correlate_message(source_xarray,receiver_xarray):
    start1 = UTCDateTime(source_xarray.attrs['starttime'])
    station_1 = list(source_xarray.coords['station_id'].values)[0]
    station_2 = list(receiver_xarray.coords['station_id'].values)[0]
    print('*' * 10)
    print('will not correlate windows!')
    print('src {} \nrec {}'.format(station_1,station_2))
    print('window: {}'.format(start1))
    print('*'*10)

def xarray_crosscorrelate(source_xarray, receiver_xarray,
                          taper=0.1, max_tau_shift=None, dummy_task=False,**kwargs):
    try:
        _check_xcorrelate_assumptions(source_xarray,receiver_xarray,taper,max_tau_shift,dummy_task,**kwargs)
    except Exception:
        _will_not_correlate_message(source_xarray,receiver_xarray)
        dummy_task=True
    src_channels   = list(source_xarray['channel'].values)
    rec_channels   = list(receiver_xarray['channel'].values)
    pair = ['src:' + list(source_xarray.coords['station_id'].values)[0] + \
            'rec:' + list(receiver_xarray.coords['station_id'].values)[0]]

    if not dummy_task:
        xcorr_np_mat = _cross_correlate_xarray_data(source_xarray,receiver_xarray,**kwargs)
    else:
        xcorr_np_mat = _dummy_correlate(source_xarray, receiver_xarray)

    xcorr_np_mat = _extract_center_of_ndarray(xcorr_np_mat, max_tau_shift, source_xarray)

    xcorr_np_mat = xcorr_np_mat.reshape((len(src_channels),len(rec_channels),1,xcorr_np_mat.shape[-1]))

    tau_array    = _get_new_time_array(source_xarray, max_tau_shift, xcorr_np_mat.shape[-1])


    xarray = xr.DataArray(xcorr_np_mat, coords=(('src_chan', src_channels),
                                                ('rec_chan', rec_channels),
                                                ('pair', pair),
                                                ('time',tau_array)))

    return xarray




def _extract_center_of_ndarray(corr_mat, tau_shift,xarray):
    if tau_shift is not None:
        center = (corr_mat.shape[-1] - 1 )//2
        shift_npts = int(tau_shift / xarray.attrs['delta'])
        corr_mat = corr_mat[:,:, center - shift_npts: center + shift_npts + 1]
    return corr_mat


def _get_new_time_array(xarray, max_tau_shift, max_samples):
    delta           = xarray.attrs['delta']
    zero            = np.datetime64(UTCDateTime(0.0).datetime)
    timedelta       = pd.Timedelta(delta, 's').to_timedelta64()
    if max_tau_shift is None:
        max_delta_shift = pd.Timedelta((max_samples-2) * delta/2 , 's').to_timedelta64()
    else:
        max_delta_shift = pd.Timedelta(max_tau_shift, 's').to_timedelta64()
    positive_tau    = np.arange(zero, zero + max_delta_shift + timedelta, timedelta)
    negative_tau    = np.arange(zero - max_delta_shift, zero,  timedelta)
    tau_array       = np.concatenate((negative_tau, positive_tau))
    return tau_array



def _cross_correlate_xarray_data(source_xarray, receiver_xarray,gpu_enable=False,torch=None,**kwargs):
    src_chan_size = source_xarray.data.shape[0]
    rec_chan_size = receiver_xarray.data.shape[0]
    time_size     = source_xarray.data.shape[-1]
    src_data      = source_xarray.data.reshape(  src_chan_size, time_size)
    receiver_data = receiver_xarray.data.reshape(rec_chan_size, time_size)


    corr_length     = time_size * 2 - 1
    target_length = fftpack.next_fast_len(corr_length)
    if not gpu_enable:

        fft_src = np.conj(np.fft.rfft(src_data, target_length, axis=-1))
        fft_rec = np.fft.rfft(receiver_data, target_length, axis=-1)

        result = _multiply_in_mat(fft_src,fft_rec)

        xcorr_mat = np.real(np.fft.fftshift(np.fft.irfft(result, corr_length, axis=-1),axes=-1)).astype(np.float64)

    else:

        zero_tensor_src = torch.zeros([target_length, src_chan_size], dtype=torch.float32, device=torch.device('cuda'))
        zero_tensor_rec = torch.zeros([target_length, rec_chan_size], dtype=torch.float32, device=torch.device('cuda'))
        zero_tensor_result= torch.zeros([target_length,src_chan_size,rec_chan_size],
                                        dtype=torch.float32, device=torch.device('cuda'))
        zero_tensor_src[:target_length,:] = np.swapaxes(src_data,1,0)
        zero_tensor_rec[:target_length,:] = np.swapaxes(receiver_data,1,0)

        fft_src  = torch.rfft(zero_tensor_src, 1)
        fft_rec  = torch.rfft(zero_tensor_src, 1)
        return None


    return xcorr_mat / time_size

def _multiply_in_mat(one,two,dtype=np.complex64):
    zero_mat = np.zeros((one.shape[0],
                         two.shape[0],
                         one.shape[-1]), dtype=dtype)

    for ri in range(0,two.shape[0]):
        zero_mat[:, ri, :]  = one

    for si in range(0,one.shape[0]):
        zero_mat[si, :, :] *= two

    return zero_mat


def _into_frequency_domain(array,axis=-1):
    target_length = fftpack.next_fast_len(array.shape[axis])
    fft           = fftpack.fft(array, target_length, axis=axis)
    return fft

def _into_time_domain(array,axis=-1):
    return np.real(fftpack.ifft(array, axis=axis,overwrite_x=True))

def _get_deltaf(array,delta,axis=-1):
    frequencies = fftpack.fftfreq(array.shape[axis], d=delta)
    return frequencies

def _dummy_correlate(source_array,
                     receiver_array):
    src_len = source_array.data.shape[0]
    rec_len = receiver_array.data.shape[0]
    time    = source_array.data.shape[-1]
    mat = np.real(_multiply_in_mat(source_array.data.reshape((src_len,time)),receiver_array.data.reshape(rec_len,time)))

    return mat


from anxcor.abstractions import XArrayProcessor
from anxcor.xarray_routines import XArrayBandpass, XArrayConverter, XArrayRemoveMeanTrend, \
    XArrayWhiten, XArrayTaper, XArrayTemporalNorm
from anxcor.core import Anxcor, AnxcorDatabase
import numpy as np

__version__='0.0.3'


import setuptools
import sys
with open("README.md", "r") as fh:
    long_description = fh.read()

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
SETUP_REQUIRES = ['pytest-runner >= 4.2'] if needs_pytest else []
TESTS_REQUIRE = ['pytest >= 2.7.1']

DISTNAME = 'anxcor'
LICENSE = 'MIT'
AUTHOR = 'Kevin A. Mendoza'
AUTHOR_EMAIL = 'kevin.mendoza@utah.edu'
URL = 'https://github.com/uofuseismo/anxcor'

setuptools.setup(
    name=DISTNAME,
    license = LICENSE,
    version="0.0.3",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description="ANXCOR is a python library for performing seismic ambient noise crosscorrelations",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url=URL,
    packages=setuptools.find_packages(),
    setup_requires=SETUP_REQUIRES,
    tests_require=TESTS_REQUIRE,
    install_requires = [
    'obspy>=1.1',
    'xarray>=0.12',
    'numpy>=1.16',
    'scipy>=1.0',
    'pandas>=0.24',
    'bottleneck>=1.2'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.6",
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering',
    ],
)
print(setuptools.find_packages())


import os
import shutil

sep = os.sep
time_format = '"D%d_M%m_Y%Y__H%H:M%M:S%S"'

def join(one,two):
    return os.path.join(one,two)


def make_path_from_list(path_list):
    path = os.path.join(*path_list)
    make_dir(path)
    return path

def make_dir(path):
    if not folder_exists(path):
        os.makedirs(path)

def get_filelist(directory):
    listOfFiles = []
    for (dirpath, dirnames, filenames) in os.walk(directory):
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
    return listOfFiles


def get_files_with_extensions(list_of_files,extensions):
    list_of_valid_files = []
    for potential_file in list_of_files:
        if potential_file.lower().endswith(extensions.lower()):
            list_of_valid_files.append(potential_file)
    return list_of_valid_files

def create_workingdir(*args,fail_if_exists=True):
    args = list(args)
    dir = args.pop(0)
    for arg in args:
        dir = dir + sep + arg
    try:
        os.makedirs(dir,exist_ok=fail_if_exists)
    except FileExistsError:
        if fail_if_exists:
            raise FileExistsError
    return dir

def delete_dirs(dirs):
    shutil.rmtree(dirs)

def delete_file(file):
    os.remove(file)

def file_exists(file):
    from os import path
    return path.exists(file) and path.isfile(file)

def folder_exists(folder):
    return os.path.isdir(folder)

def _clean_dirs_of_index(dirlist):
    for dir in dirlist:
        try:
            os.remove(dir + './index.h5')
        except Exception:
            pass

def _clean_files_in_dir(dir):
    filelist = get_filelist(dir)
    for dir in filelist:
        try:
            os.remove(dir)
        except Exception:
            pass

def _clean_dirs_and_files(dirlist):
    for dir in dirlist:
        try:
            os.remove(dir)
        except Exception:
            pass
        try:
            shutil.rmtree(dir)
        except Exception:
            pass

def _how_many_fmt(directory,format='.sac'):
    filelist = get_filelist(directory)
    saclist  = get_files_with_extensions(filelist, format)
    return len(saclist)

def get_folderpath(filepath):

    for i in range(len(filepath)-1,-1,-1):
        if filepath[i]==os.sep:
            folder = filepath[:i]
            return folder


import  anxcor.utils as utils
from obspy.core import UTCDateTime
import xarray as xr
import sys
import json

def write(xarray, path, extension):
    array_path      = '{}{}{}{}'.format(path, utils.sep, extension, '.nc')
    attributes_path = '{}{}{}{}'.format(path, utils.sep, extension, '.metadata.json')
    data = xarray.copy()
    data.attrs = {}
    data.to_netcdf(array_path)
    with open(attributes_path, 'w') as p_file:
        json.dump(xarray.attrs, p_file, sort_keys=True, indent=4)


def read(path, extension):
    xarray_path     ='{}{}{}{}'.format(path, utils.sep, extension, '.nc')
    attributes_path ='{}{}{}{}'.format(path, utils.sep, extension, '.metadata.json')
    try:
        xarray = xr.open_dataset(xarray_path)
        with open(attributes_path, 'r') as p_file:
            attrs = json.load(p_file)
    except FileNotFoundError:
        print('File:\n {}\n not found. Ignoring window'.format(attributes_path))
        return None
    xarray.attrs = attrs
    return xarray


class _IO:

    def __init__(self, dir):
        self._file      = dir
        self._isenabled = False

    def enable(self):
        self._isenabled=True

    def is_enabled(self):
        return self._isenabled

    def get_folder_extension(self, xarray):
        if 'stacks' in xarray.attrs.keys():
            # then xarray represents stack data
            pair = list(xarray.coords['pair'].values)[0]
            stack_num = str(xarray.attrs['stacks'])
            return self._file + utils.sep + pair + utils.sep + stack_num
        elif 'operations' in xarray.attrs.keys():
            # then xarray represents single station data
            operation = xarray.attrs['operations'].split('\n')[-1]
            starttime = UTCDateTime(xarray.attrs['starttime']).isoformat()
            return self._file + utils.sep + operation + utils.sep + starttime
        else:
            # then xarray is a dataset
            pair_list = list(xarray.coords['pair'].values)
            strsum = pair_list[0] + '|' + str(len(pair_list))
            return self._file + utils.sep + strsum

    def get_filename(self, xarray):
        if 'stacks' in xarray.attrs.keys():
            # then xarray represents stack data
            starttime = UTCDateTime(xarray.attrs['starttime']).isoformat()
            return starttime
        elif 'operations' in xarray.attrs.keys():
            # then xarray represents single station data
            station = list(xarray.coords['station_id'].values)[0]
            return station
        else:
            return 'combined_data'


class _XArrayWrite(_IO):

    def __init__(self, directory=None):
        super().__init__(dir)

    def set_folder(self, file):
        self.enable()
        if not utils.folder_exists(file):
            utils.make_dir(file)
        self._file = file

    def _chkmkdir(self,dir):
        if not utils.folder_exists(dir):
            utils.make_dir(dir)

    def __call__(self, xarray, process, folder, file, dask_client=None, **kwargs):
        if self._file is not None and xarray is not None:
            folder    = '{}{}{}{}{}'.format(self._file, utils.sep, process, utils.sep, folder)
            self._chkmkdir(folder)
            write(xarray, folder, file)
        return None


class _XArrayRead(_IO):

    def __init__(self, directory=None):
        super().__init__(directory)
        self._file = directory

    def set_folder(self, directory):
        self.enable()
        if not utils.folder_exists(directory):
            utils.make_dir(directory)
        self._file = directory

    def __call__(self, xarray, process, folder, file, dask_client=None, **kwargs):
        result = None
        if self._file is not None:
            folder = self._file + utils.sep + process + utils.sep + folder
            result = read(folder, file)

        if result is not None:
            return result
        else:
            return xarray

class _XDaskTask:

    def __init__(self,dummy_task=False,**kwargs):
        self._kwargs = kwargs
        self._kwargs['dummy_task']=dummy_task
        self.read  = _XArrayRead(None)
        self.write = _XArrayWrite(None)
        self._enabled = True

    def disable(self):
        self._enabled=False

    def set_io_task(self, folder, action, **kwargs):
        if action=='save':
            self.write.set_folder(folder)
        else:
            self.read.set_folder(folder)


    def set_kwargs(self, kwarg):
        for key, value in kwarg.items():
            if key in self._kwargs.keys():
                self._kwargs[key]=value
            else:
                print('key [{}] is not a assignable parameter for {}\n'.format(key,self.get_name())+\
                    'skipping...')

    def get_kwargs(self):
        return {**self._kwargs}

    def get_name(self):
        return self._get_process()

    def __call__(self, *args, dask_client=None, **kwargs):
        key = self._get_operation_key(**kwargs)

        result = None
        if self._enabled and not self.read.is_enabled() and self._should_process(*args):
            if dask_client is None:
                result = self._execute(*args,**kwargs)
            else:
                result = dask_client.submit(self._execute, *args, key=key,**kwargs)

        result = self._io_operations(*args, dask_client=dask_client, result=result,**kwargs)
        return result

    def _execute(self, *args, **kwargs):
        persist_name       = self.__get_name(*args)
        persisted_metadata = self.__metadata_to_persist(*args, **kwargs)
        if args is None or len(args)==1 and args[0] is None:
            result = None
        else:
            result = self._single_thread_execute(*args, **kwargs)
        self._assign_metadata(persist_name, persisted_metadata, result)
        return result

    def _assign_metadata(self, persist_name, persisted_metadata, result):
        if result is None:
            return None
        if persisted_metadata is not None:
            result.attrs = persisted_metadata
        if persist_name is not None:
            result.name = persist_name

    def _io_operations(self, *args, dask_client=None, result=None, starttime=0, station=0,**kwargs):
        key = self._get_operation_key(starttime,station)
        if self.read.is_enabled():
            file, folder, process = self._get_io_string_vars(starttime, station)
            if dask_client is None:
                result = self._read_execute(result, process, folder, file)
            else:
                result = dask_client.submit(self._read_execute,*args, process, folder, file,key=key)
        elif self.write.is_enabled():
            file, folder, process = self._get_io_string_vars(starttime, station)
            if dask_client is None:
                self.write(result, process, folder, file)
            else:
                if 'dask' not in sys.modules:
                    from dask.distributed import fire_and_forget
                end = dask_client.submit(self.write, result,  process, folder, file, key='writing: '+ key)
                fire_and_forget(end)
        return result

    def _get_io_string_vars(self, starttime, station):
        process = self._get_process()
        folder = self._window_key_convert(starttime)
        file = station
        return file, folder, process

    def _read_execute(self, result, process, folder, file):
        result = self.read(result, process,folder,file)
        result = self._addition_read_processing(result)
        return result

    def _single_thread_execute(self,*args,**kwargs):
        pass

    def _dask_task_execute(self,*args,**kwargs):
        pass

    def __metadata_to_persist(self,*param,**kwargs):
        if param is None or (len(param)==1 and param[0] is None):
            return None
        else:
            return self._metadata_to_persist(*param,**kwargs)

    def __get_name(self,*param,**kwargs):
        if param is None or (len(param)==1 and param[0] is None):
            return None
        else:
            return self._get_name(*param,**kwargs)


    def _metadata_to_persist(self, *param, **kwargs):
        if len(param)==1:
            attrs = param[0].attrs.copy()
        else:
            attrs = {**param[0].attrs.copy(), **param[1].attrs.copy()}
        added_kv_metadata = self._add_metadata_key()
        add_operation     = self._add_operation_string()
        if added_kv_metadata is not None:
            if isinstance(added_kv_metadata[0],tuple) or isinstance(added_kv_metadata[0],list):
                for key, value in added_kv_metadata:
                    attrs[added_kv_metadata[0]] = added_kv_metadata[1]
            else:
                attrs[added_kv_metadata[0]] = added_kv_metadata[1]
        if add_operation is not None:
            attrs['operations']=attrs['operations'] + '\n' + add_operation
        return attrs

    def _get_name(self,*args,**kwargs):
        if len(args) == 1:
            return args[0].name
        elif args[0] is None and args[1] is not None:
            return args[1].name
        elif args[0] is not None and args[1] is None:
            return args[0].name
        else:
            return args[0].name + ':' + args[1].name

    def _add_operation_string(self):
        return None

    def _add_metadata_key(self):
        return None

    def _get_process(self):
        return 'process'

    def _addition_read_processing(self, result):
        return result

    def _window_key_convert(self,window):
        return window

    def _get_operation_key(self,starttime=0,station=0,**kwargs):
        window_key = self._window_key_convert(starttime)
        return '{}-{}-{}'.format(self._get_process(),station,window_key)

    def _should_process(self, *args):
        return True


class XArrayProcessor(_XDaskTask):

    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

    def _addition_read_processing(self, result):
        if result is not None:
            name   = list(result.data_vars)[0]
            xarray       = result[name].copy()
            xarray.attrs = result.attrs.copy()
            del result
            return xarray
        return result

    def  _should_process(self,xarray, *args):
        return xarray is not None

    def _window_key_convert(self,window):
        return UTCDateTime(int(window*100)/100).isoformat()


class XDatasetProcessor(_XDaskTask):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _addition_read_processing(self, result):
        return result

